//
//  XpcUtility.c
//  RecoveRx
//
//  Created by Sendoh Chen on 2014/7/16.
//
//

#include <pthread.h>
#include <syslog.h>
#include "XpcUtility.h"
#include "RecoveryDef.h"
#include <TestDisk/TestDiskController.h>

xpc_connection_t remote;
sRecArg recArg;

void __createStartRecoveryThread(sRecArg *recArg);
void __XPC_Send_Status_Message(sRecStatus *recStatus);
void __XPC_Send_Finish_Message(recovery_pstatus_t status);

void __XPC_Peer_Event_Handler(xpc_connection_t connection, xpc_object_t event)
{
	xpc_type_t type = xpc_get_type(event);
    
	if (type == XPC_TYPE_ERROR) {
		if (event == XPC_ERROR_CONNECTION_INVALID) {
			// The client process on the other end of the connection has either
			// crashed or cancelled the connection. After receiving this error,
			// the connection is in an invalid state, and you do not need to
			// call xpc_connection_cancel(). Just tear down any associated state
			// here.
            
		} else if (event == XPC_ERROR_TERMINATION_IMMINENT) {
			// Handle per-connection termination cleanup.
		}
        
	} else if (type == XPC_TYPE_DICTIONARY) {
        remote = xpc_dictionary_get_remote_connection(event);
        
        // Parse request and params
        const char *request = xpc_dictionary_get_string(event, RECOVERY_REQUEST);
        if (request && strcmp(request, RECOVERY_START) == 0) {
            syslog(LOG_NOTICE, "RECOVERY_START");
            
            // Get Disk Name
            strncpy(recArg.diskName, xpc_dictionary_get_string(event, RECOVERY_DISK_NAME), MAX_DISK_NAME_LEN);
            syslog(LOG_NOTICE, "Disk Name: %s", recArg.diskName);
            
            // Get Partition Number
            recArg.partNum = (int)xpc_dictionary_get_int64(event, RECOVERY_PART_NUM);
            syslog(LOG_NOTICE, "Partition Number: %d", recArg.partNum);
            
            // Get Path
            strncpy(recArg.path, xpc_dictionary_get_string(event, RECOVERY_PATH), MAX_PATH_LEN);
            syslog(LOG_NOTICE, "Path: %s", recArg.path);
            
            // Get Type
            xpc_object_t types = xpc_dictionary_get_value(event, RECOVERY_TYPE);
            recArg.typesCount = (int)xpc_array_get_count(types);
            for (int i = 0; i < recArg.typesCount; i++) {
                const char *type = xpc_array_get_string(types, i);
                if (type) {
                    //syslog(LOG_NOTICE, "Type: %s", type);
                    strncpy(recArg.types[i], type, MAX_FILE_TYPE_STR_LEN);
                }
            }
            
            // Create Start Recovery Thread
            __createStartRecoveryThread(&recArg);
            
            // Send response
            xpc_object_t response = xpc_dictionary_create_reply(event);
            xpc_dictionary_set_string(response, RECOVERY_RESPONSE, RECOVERY_RESPONSE_SUCCESS);
            xpc_connection_send_message(remote, response);
            xpc_release(response);
        }
        else if (request && strcmp(request, RECOVERY_STOP) == 0) {
            syslog(LOG_NOTICE, "RECOVERY_STOP");
            
            stopRecovery();
            
            //Send response
            xpc_object_t response = xpc_dictionary_create_reply(event);
            xpc_dictionary_set_string(response, RECOVERY_RESPONSE, RECOVERY_RESPONSE_SUCCESS);
            xpc_connection_send_message(remote, response);
            xpc_release(response);
        }
        else if (request && strcmp(request, RECOVERY_EXIT) == 0) {
            syslog(LOG_NOTICE, "RECOVERY_EXIT");
            exit( 0 );
        }
	}
}

void* __Start_Recovery_Thread(void *data)
{
    sRecArg *rec_arg = (sRecArg *)data;
    
    startRecovery(rec_arg->diskName, rec_arg->partNum, rec_arg->path, rec_arg->types, rec_arg->typesCount,
                  __XPC_Send_Status_Message,
                  __XPC_Send_Finish_Message);

    return NULL;
}

void __createStartRecoveryThread(sRecArg *rec_arg)
{
    pthread_t tid;
    int err = pthread_create(&tid, NULL, __Start_Recovery_Thread, (void *)rec_arg);
    if (err != 0)
        syslog(LOG_NOTICE, "Can't create thread :[%s]", strerror(err));
    else
        syslog(LOG_NOTICE, "Thread created successfully");
}

void __XPC_Connection_Handler(xpc_connection_t connection)
{
    syslog(LOG_NOTICE, "Configuring message event handler for RecoveryHelper");

	xpc_connection_set_event_handler(connection, ^(xpc_object_t event) {
		__XPC_Peer_Event_Handler(connection, event);
	});
	
	xpc_connection_resume(connection);
}

void __XPC_Send_Status_Message(sRecStatus *recStatus)
{
    xpc_object_t message = xpc_dictionary_create(NULL, NULL, 0);
    if (strlen(recStatus->progress))
        xpc_dictionary_set_string(message, RECOVERY_PROGRESS, recStatus->progress);
    if (strlen(recStatus->sectorRead))
        xpc_dictionary_set_string(message, RECOVERY_SECTOR, recStatus->sectorRead);
    if (strlen(recStatus->elapsedTime))
        xpc_dictionary_set_string(message, RECOVERY_ELAPSEDTIME, recStatus->elapsedTime);
    if (strlen(recStatus->estimatedTime))
        xpc_dictionary_set_string(message, RECOVERY_ESTIMATEDTIMELEFT, recStatus->estimatedTime);
    if (strlen(recStatus->fileFound))
        xpc_dictionary_set_string(message, RECOVERY_FILEFOUND, recStatus->fileFound);
    
    if (recStatus->fileRecoveredCount) {
        xpc_object_t fileRecoveredStr = xpc_array_create(NULL, 0);
        xpc_object_t fileRecoveredNum = xpc_array_create(NULL, 0);
        for (int i = 0; i < recStatus->fileRecoveredCount; i++) {
            xpc_array_append_value(fileRecoveredStr, xpc_string_create(recStatus->fileRecoveredStr[i]));
            xpc_array_append_value(fileRecoveredNum, xpc_int64_create(recStatus->fileRecoveredNum[i]));
        }

        xpc_dictionary_set_value(message, RECOVERY_FILERECOVEREDSTR, fileRecoveredStr);
        xpc_dictionary_set_value(message, RECOVERY_FILERECOVEREDNUM, fileRecoveredNum);
    }
    
    xpc_connection_send_message(remote, message);
}

void __XPC_Send_Finish_Message(recovery_pstatus_t status)
{
    xpc_object_t message = xpc_dictionary_create(NULL, NULL, 0);
    
    switch(status)
    {
        case RECOVERY_PSTATUS_OK:
            syslog(LOG_NOTICE, "Recovery completed.");
            xpc_dictionary_set_string(message, RECOVERY_FINISH, RECOVERY_FINISH_OK);
            break;
        case RECOVERY_PSTATUS_STOP:
            syslog(LOG_NOTICE, "Recovery aborted by the user.");
            xpc_dictionary_set_string(message, RECOVERY_FINISH, RECOVERY_FINISH_STOP);
            break;
        case RECOVERY_PSTATUS_EACCES:
            syslog(LOG_NOTICE, "Cannot create file in current directory.");
            xpc_dictionary_set_string(message, RECOVERY_FINISH, RECOVERY_FINISH_EACCES);
            break;
        case RECOVERY_PSTATUS_ENOSPC:
            syslog(LOG_NOTICE, "Cannot write file, no space left.");
            xpc_dictionary_set_string(message, RECOVERY_FINISH, RECOVERY_FINISH_ENOSPC);
            break;
        default:
            syslog(LOG_NOTICE, "Recovery completed.");
            xpc_dictionary_set_string(message, RECOVERY_FINISH, RECOVERY_FINISH_OK);
            break;
    }
    
    xpc_connection_send_message(remote, message);
}
